{{
SpinStruct: malloc utilities
}}

OBJ
  vga: "VGA_Text"
  m  : "malloc"
  
PUB dump(label) | ii
{
/**
 * dump the heap
 */
 _ void dump();
}
  result := m.getHeapStart
  repeat while (m.isNotEnd(result)) '(m.getNextPtr(result) < m.getHeapEnd and m.getNextPtr(result) <> 0)
    printentry(result,label)
    result := m.getNextPtr(result)
    
PUB printentry(ptr,stype)    
{
/**
 * print a heap entry's critical data
 */
 _ void printentry(void* heapptr, char* stype);
}
  'print(string("dump"), result, nextptr(result))
  vga.str(string($d,"P"))
  vga.str(stype)
  vga.hex(ptr, m.getSizeofInt<<1)
  'vga.hex(getvalue(ptr), m.getSizeofInt<<1)
  if m.isFree(ptr)
    vga.str(string(" a "))
  else
    vga.str(string(" u "))
  vga.hex(ptr, m.getSizeofInt<<1)
  vga.str(string("->"))
  vga.hex(m.getNextPtr(ptr), m.getSizeofInt<<1)
  vga.out(" ")
  vga.hex(m.getNextPtr(m.getNextPtr(ptr)), m.getSizeofInt<<1)
  vga.str(string(" # "))
  vga.dec(length(m.getUserPtr(ptr)))

PUB showHeap
{
/**
 * Show heap statistics
 */
 _ void showHeap();
}
  vga.str(string($d,"H#"))
  vga.dec(getHeapCount)
  vga.str(string(" Use "))
  vga.dec(getHeapUsed)
  vga.str(string(" Free "))
  vga.dec(getHeapFree)
  vga.str(string(" Bytes "))
  vga.dec(m.getHeapEnd-m.getHeapStart) 'total length is num bytes used + size of heapptr
  'vga.out($d)

PUB print(fn, ptr, val)
{
/**
 * print string and two variables in hex
 */
 _ void print(char* fn, int ptr, int val);
}
  vga.out($d)
  vga.str(fn)
  vga.out($9)
  vga.hex(ptr, (m.getSizeofInt<<1))
  vga.out(":")
  vga.hex(val, (m.getSizeofInt<<1))

PUB printb(buf) | ii, len
  len := length(buf)
  ii := 0
  repeat while(byte[buf+ii]<>0 and ii < len)
    vga.out(byte[buf+ii++])


PUB length(ptr)
{{
/**
 * get allocated length in bytes of pointer
 * @param ptr is pointer returned by malloc
 * @returns 0 if pointer not found, else length in bytes allocated
 */
 _ int length(void* ptr);
}}
  if m.isptr(m.heapptr(ptr)) == 0
    return 0
  return m.nextptr(m.heapptr(ptr)) - m.heapptr(ptr) - m.getsizeofint

PUB getHeapCount | p
{{
/**
 * Get number of heap entries
 * @returns number of heap entries
 */
 _ int getHeapUsed();
}}
  result := 0
  p := m.getHeapStart
  repeat while (m.nextptr(p) < m.getHeapEnd)
    result ++
    p := m.nextptr(p)      

PUB getHeapFree | p
{{
/**
 * Get bytecount of heap free
 * @returns bytes free
 */
 _ int getHeapUsed();
}}
  result := 0
  p := m.getHeapStart
  repeat while (m.nextptr(p) < m.getHeapEnd)
    if m.isFree(p)
      result += length(m.block(p))
    p := m.nextptr(p)      

PUB getHeapUsed | p
{{
/**
 * Get bytecount of heap in use
 * @returns bytes in use
 */
 _ int getHeapUsed();
}}
  result := 0
  p := m.getHeapStart
  repeat while (m.nextptr(p) < m.getHeapEnd)
    if m.isFree(p) == 0
      result += length(m.block(p))      
    p := m.nextptr(p)

PUB getHeapBytes | p
{{
/**
 * Get bytecount of heap
 * @returns number of user bytes in heap
 */
 _ int getHeapBytes();
}}
  result := m.getHeapEnd - m.getHeapStart
{
  p := heapStart
  repeat while (nextptr(p) < heapEnd and nextptr(p) <> 0)
    result += length(block(p))+sizeof_int      
    p := nextptr(p)
}

