{{
SpinStruct: Singly Linked List

This implements a singly linked list that grows "up" or "down"
depending on the addItem or insertItem method being used.

A "long" .spin node:

VAR
  long data
  long next

  Node traversal:
    ptr := data
    ptr := long[ptr+sizeof(int)] ' using ptr := getNext(ptr) encapsulates details

A comparable C struct node:

  typedef struct ns {
          int data;
          struct ns *next;
  } node_t;
  node_t top, ptr;   

  Node traversal:
    ptr = top
    ptr = ptr->next

List works like this:

no item list:  *first = NULL,  *first->next = NULL

 1 item list:  *ptr1 = data1, *ptr1->next = NULL
               *first  = ptr1
 
 2 item list:  *ptr1 = data1, *ptr1->next = NULL
               *ptr2 = data2, *ptr2->next = ptr1
               *first  = ptr2
               
 3 item list:  *ptr1 = data1, *ptr1->next = NULL
               *ptr2 = data2, *ptr2->next = ptr1
               *ptr3 = data3, *ptr3->next = ptr2
               *first  = ptr3

 N item list:  *ptrN = dataN, *nextN = *nextP (with nextP = previously allocated pointer)
               *first  = ptrN
}}

CON  NULL = 0
OBJ  m: "malloc" ' memory manager
 
VAR
' typedef struct ns {
'   int data;
'   struct ns *next;
' } node;
  long first ' "first" of list is first item in the list
  long nxt1  ' the "next" pointer - only the space is used not the name
  long last  ' value contains "last" known added pointer

{
  List Traversal:     
  p := top
  repeat while(getNext(p) <> 0)
    printItem(p)
    p := getNext(p)
}

PUB init
{{ init list }}
  'first := $F151
  last  := 0                    ' last is a placeholder for previously added node
  setData(@first, 0)            ' set first value
  setNext(@first, 0)            ' set next null

PUB top
{{ returns first pointer }}
  result := first

PUB end
{{ returns last known pointer }}
  result := last

PUB addItem(val) | p, ptr
{{ add value as ptr }}
  ptr := m.malloc(2*m.getSizeOfInt)                     ' get storage for new item
  if top == 0          ' if first == 0, make this ptr first
    setData(ptr, val)
    setNext(ptr, 0)
    first := ptr
  else
    if last <> 0
      p := last
    setNext(p,ptr)              ' set p->next = ptr
    setData(ptr, val)           ' save new item
    setNext(ptr, 0)             ' set ptr->next = 0 to terminate list
  last := ptr                   ' finally, save this ptr as last
  result := ptr
  
PUB insertItem(entry, val) | p, ptr
{{ insert value before entry }}
  if entry == top
    ptr := m.malloc(2*m.getSizeOfInt) ' get storage for new item
    setNext(ptr, first)
    setData(ptr, val)
    first := ptr
  else
    p := findNext(entry)
    if p == 0
      return 0
    ptr := m.malloc(2*m.getSizeOfInt) ' get storage for new item
    setNext(p, ptr)
    setData(ptr, val)             ' save new item
    setNext(ptr, entry)           ' set ptr next = entry next
  result := ptr

PUB setItem(ptr, val)
{{ add value as ptr }}
  if findEntry(ptr) == 0
    return
  setData(ptr, val)             ' set item
  result := ptr
  'dump
  'printItem(ptr)
  
PUB getItem(ptr)
{{ get value from ptr }}
  case m.getSizeOfInt
    2: result := word[ptr]
    4: result := long[ptr]

PUB deleteItem(ptr) | p, n
  if ptr == 0
    return
  if ptr == first
    first := getNext(ptr)
    m.free(ptr)
  else
    n := findNext(ptr)
    if n <> 0
      setNext(n, getNext(ptr))
      m.free(ptr)
    
PUB getData(ptr)
  case m.getSizeOfInt
    2: result := word[ptr]
    4: result := long[ptr]

PUB setData(ptr, val)
  case m.getSizeOfInt
    2: word[ptr] := val
    4: long[ptr] := val

PUB getNext(ptr)
  case m.getSizeOfInt
    2: result := word[ptr+m.getSizeOfInt]
    4: result := long[ptr+m.getSizeOfInt]

PUB setNext(ptr, val)
  case m.getSizeOfInt
    2: word[ptr+m.getSizeOfInt] := val
    4: long[ptr+m.getSizeOfInt] := val
      
PUB delete | ptr, p
{{ remove list and all pointers }}
  ptr := first
  repeat while(ptr <> 0)
    p := ptr
    ptr := getNext(ptr)
    m.free(p)
    
PUB findData(val) | p
{{ find pointer with val in list }}
  p := first
  repeat while(p <> 0)
    if getData(p) == val
      return p
    p := getNext(p)
  result := 0    

PUB findEntry(ptr) | p
{{ find pointer with ptr = next in list }}
  p := first
  repeat while(p <> 0)
    if p == ptr
      return p
    p := getNext(p)
  result := 0    

PUB findNext(ptr) | p
{{ find pointer with ptr = next in list }}
  p := first
  repeat while(p <> 0)
    if getNext(p) == ptr
      return p
    p := getNext(p)
  result := 0    


{
OBJ   vga: "VGA_Text"
mu    : "malloc_utils"

PRI print2(v1, v2)
  vga.out($d)
  vga.hex(v1,4)
  vga.out($20)
  vga.hex(v2,4)

PRI print4(v1, v2, v3, v4)
  vga.out($d)
  vga.hex(v1,4)
  vga.out($20)
  vga.hex(v2,4)
  vga.out($20)
  vga.hex(v3,4)
  vga.out($20)
  vga.hex(v4,4)

PUB printItem(ptr)
  print4(ptr, getData(ptr), getNext(ptr), $BEEF)
}
   

